!==================================================================
!Interface to cusolverDn C functions
!==================================================================
    
  module cu_solver

  use iso_c_binding
  
  interface
     
     integer (c_int) function cusolverDnCreate(cusolver_Hndl) bind(C,name="cusolverDnCreate")
     
     use iso_c_binding
     implicit none
     
     type(c_ptr) cusolver_Hndl
     
     end function
  end interface  
  
  interface
     
     integer (c_int) function cusolverDnDestroy(cusolver_Hndl) bind(C,name="cusolverDnDestroy")
     
     use iso_c_binding
     implicit none
     
     type(c_ptr) cusolver_Hndl
     
     end function
  end interface 
  
  interface
    integer(c_int) function cusolverDnSgetrf_bufferSize(cusolver_Hndl,m,n,d_A,lda,Lwork) bind(C,name="cusolverDnSgetrf_bufferSize") 
      
      use iso_c_binding
      implicit none
    
      type(c_ptr) cusolver_Hndl
      integer(c_int),value::m
      integer(c_int),value::n
      type(c_ptr)  d_A
      integer(c_int),value::lda
      integer(c_int)::Lwork
    end function
  end interface 
                                
  interface
     
     integer (c_int) function cusolverDnSgetrf(cusolver_Hndl,m,n,d_A,lda,d_WS,d_Ipiv,d_devInfo) bind(C, name="cusolverDnSgetrf")
       use iso_c_binding
       implicit none
       type(c_ptr) cusolver_Hndl
       integer(c_int),value:: m
       integer(c_int),value::n
       type(c_ptr) d_A
       integer(c_int),value::lda
       type(c_ptr) d_WS
       type(c_ptr) d_Ipiv
       type(c_ptr) d_devInfo
       
     end function 
    
  end interface
  
  interface
     
     integer (c_int) function cusolverDnSgetrs(cusolver_Hndl,trans,n,nrhs,d_A,lda,d_Ipiv,d_B,ldb,d_devInfo) bind(C, name="cusolverDnSgetrs")
       use iso_c_binding
       implicit none
       type(c_ptr) cusolver_Hndl
       character(kind=c_char) trans
       integer(c_int), value::n
       integer(c_int), value::nrhs
       type(c_ptr) d_A
       integer(c_int), value::lda    
       type(c_ptr) d_Ipiv
       type(c_ptr) d_B
       integer(c_int), value::ldb
       type(c_ptr) d_devInfo
       
     end function cusolverDnSgetrs
   
  end interface  
  
end module 

!=========================================
program main
!=========================================

! The purpose of this routine is to provide for GPU based solution of 
! large dense systems of equations in legacy FORTRAN Applications.

! This is the development version of the routine which does not yet work ...

  use iso_c_binding
  use cuda_h
  use cuda_runtime_h
  use cu_solver
  
  
  ! ------------- matrix definition & host CPU storage variables 
  
  integer(c_int) m,n,lda,ldb
  
  real, allocatable :: A(:,:),ATest(:,:)
  real, allocatable :: B(:,:)
  real, allocatable :: X(:,:)
  integer, allocatable::Ipiv(:)
  integer, allocatable::Lwork
  
!-------------------- CPU equivalents of device variables
  integer workspace
  integer devInfo
  character(c_char) trans
  integer(c_int) nrhs
  
  integer A_size,B_size,X_size,Ipiv_size,devInfo_size,Lwork_size
  
! -------------------- handle to device
  
  type(c_ptr) :: cusolver_Hndl

  
! -------------------- pointers to device memory
  
  type(c_ptr)::d_A
  type(c_ptr)::d_B
  type(c_ptr)::d_Lwork
  type(c_ptr)::d_WS
  type(c_ptr)::d_Ipiv
  type(c_ptr)::d_devInfo
  
! --------------------- function result variables
  
  integer cusolver_stat
  integer A_mem_stat
  integer B_mem_stat
  integer X_mem_stat
  integer Ipiv_mem_stat
  integer WS_mem_stat
  integer devInfo_mem_stat
  integer Lwork_mem_stat
  
! -------------------- pointers to host CPU memory  
  
  type(c_ptr)::CPU_A_ptr
  type(c_ptr)::CPU_ATest_ptr
  type(c_ptr)::CPU_B_ptr
  type(c_ptr)::CPU_X_ptr
  type(c_ptr)::CPU_Lwork_ptr
    
  target A,ATest,B,X,Ipiv,Lwork
  
  integer*8 free,total
  integer res
  
  ! ================================================
  free = 0
  total = 0
  res = 1
  
  res = cudaMemGetInfo(free,total)
  
  lda = 3
  ldb = 3
  m=lda
  n=lda
  nrhs = 1
  
  allocate(A(lda,n))
  allocate(ATest(lda,n))
  allocate(B(n,1))
  allocate(X(n,1))
  allocate(Ipiv(lda))
  allocate(Lwork)
  
  A_size = SIZEOF(A)
  B_size = SIZEOF(B)
  X_size = SIZEOF(X)
  Ipiv_size = SIZEOF(Ipiv)
  devInfo_size = SIZEOF(devInfo)
  Lwork_size = SIZEOF(Lwork)
  
  !    EXAMPLE FROM CUSOLVER_Library.pdf  
  !  "QR FACTORIZATION DENSE LINEAR SOLVER" 
  
  ! -------------- DEFINE [A] AND [B] --------------------------
  
  !           [A]         [x]  =    [B]
  !    | 1.0  2.0  3.0 | |1.0|    | 6.0|
  !    | 4.0  5.0  6.0 | |1.0| =  |15.0|
  !    | 2.0  1.0  1.0 | |1.0|    | 4.0|
  
  A(1,1) = 1.0;   A(1,2) = 2.0;  A(1,3) = 3.0
  A(2,1) = 4.0;   A(2,2) = 5.0;  A(2,3) = 6.0
  A(3,1) = 2.0;   A(3,2) = 1.0;  A(3,3) = 1.0
  
  B(1,1) = 6.0
  B(2,1) = 15.0
  B(3,1) = 4.0
  
  CPU_A_ptr = C_LOC(A)
  CPU_ATest_ptr = C_LOC(ATest)
  CPU_B_ptr = C_LOC(B)
  CPU_X_ptr = C_LOC(X)
  CPU_Lwork_ptr = C_Loc(Lwork)
  
! Step 1: Create cudense handle ---------------
  
  cusolver_stat = cusolverDnCreate(cusolver_Hndl)  
      
! Step 2: copy A and B to Device
  
  A_mem_stat    = cudaMalloc(d_A,A_size)
  B_mem_stat    = cudaMalloc(d_B,B_size)  
  
! ---------- also allocate space for other device based variables  
  
  Ipiv_mem_stat = cudaMalloc(d_Ipiv,Ipiv_size)
  devInfo_mem_stat = cudaMalloc(d_devInfo,devInfo_size)
  Lwork_mem_stat   = cudaMalloc(d_Lwork,Lwork_size)
  
! ---------- copy A and B to Device
  
  A_mem_stat = cudaMemcpy(d_A,CPU_A_ptr,A_size,cudaMemcpyHostToDevice)
  B_mem_stat = cudaMemcpy(d_B,CPU_B_ptr,B_size,cudaMemcpyHostToDevice)
  
! ---------- check transfer of A matrix by copy back to [ATest]
  
  A_mem_stat = cudaMemcpy(CPU_ATest_ptr,d_A,A_size,cudaMemcpyDeviceToHost)  ! Test A matrix memory copy
  
! Step 3: query working space of Sgetrf (and allocate memory on device)
  
  Lwork = 0  
  cusolver_stat = cusolverDnSgetrf_bufferSize(cusolver_Hndl,m,n,d_A,lda,Lwork) 
  
  Workspace = 4*Lwork
  WS_mem_stat = cudaMalloc(d_WS,Workspace)
  
! Step 4: compute LU factorization of [A] 
! A Memory acess error occurs here...
! Access violation reading location 0xffffffffffffffff.
    
  cusolver_stat = cusolverDnSgetrf(cusolver_Hndl,m,n,d_A,lda,d_WS,d_Ipiv,d_devInfo) 
! Step 5: compute solution vector [X] for Right hand side [B]
  
  trans = 'N'
  cusolver_stat = cusolverDnSgetrs(cusolver_Hndl,trans,n,nrhs,d_A,lda,d_Ipiv,d_B,ldb,d_devInfo)  
  
! Step 6: copy solution vector stored in [B] on device into [X] vector on host
  
  X_mem_stat = cudaMemcpy(CPU_X_ptr,d_B,B_size,cudaMemcpyDeviceToHost)  
  
! step 7: free memory on device and release CPU-side resources
  
  A_mem_Stat    = cudafree(d_A)
  B_mem_Stat    = cudafree(d_B)
  Ipiv_mem_stat = cudafree(d_Ipiv)
  WS_mem_stat   = cudafree(d_WS)
  Lwork_mem_stat = cudafree(d_Lwork)
  
  cusolver_stat = cusolverDnDestroy(cusolver_Hndl)
  
  ! Step 8: deallocate memory on host before exit
  
  deallocate(A)
  deallocate(ATest)
  deallocate(B)
  deallocate(X)
  deallocate(Ipiv)
  deallocate(Lwork)

end program 